#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 15 08:36:26 2024

@author: mike
"""
import logging
import uuid6 as uuid
import urllib3
import booklet
from typing import Union
import s3func
import weakref
import orjson
import base64
import datetime
import concurrent.futures

logger = logging.getLogger(__name__)
from . import utils


###############################################
### Parameters

ebooklet_types = ('EVariableLengthValue', 'RemoteConnGroup')


###############################################
### Functions


def check_remote_conn(remote_conn, flag):
    """

    """
    if isinstance(remote_conn, str):
        if flag != 'r':
            raise ValueError('If remote_conn is a url string, then flag must be r.')
        remote_conn = S3Connection(db_url=remote_conn)
    elif isinstance(remote_conn, dict):
        if 'remote_conn' in remote_conn:
            remote_conn = S3Connection(**remote_conn['remote_conn'])
        else:
            remote_conn = S3Connection(**remote_conn)
    elif not isinstance(remote_conn, S3Connection):
        raise TypeError('remote_conn must be either a url string or a remote.S3Connection.')

    return remote_conn


def check_write_config(
        access_key_id: str=None,
        access_key: str=None,
        db_key: str=None,
        bucket: str=None,
        endpoint_url: str=None,
        ):
    """

    """
    if isinstance(access_key_id, str) and isinstance(access_key, str) and isinstance(db_key, str) and isinstance(bucket, str):
        if isinstance(endpoint_url, str):
            if not s3func.utils.is_url(endpoint_url):
                raise TypeError(f'{endpoint_url} is not a proper url.')
        return True
    return False


def create_s3_read_session(
        access_key_id: str=None,
        access_key: str=None,
        db_key: str=None,
        bucket: str=None,
        endpoint_url: str=None,
        db_url: str=None,
        threads: int=20,
        read_timeout: int=60,
        retries: int=3,
        ):
    """

    """
    if isinstance(db_url, str):
        if not s3func.utils.is_url(db_url):
            raise TypeError(f'{db_url} is not a proper url.')
        read_session = s3func.HttpSession(max_connections=threads, read_timeout=read_timeout, stream=False, max_attempts=retries)
        key = db_url
    elif isinstance(access_key_id, str) and isinstance(access_key, str) and isinstance(db_key, str) and isinstance(bucket, str):
        if isinstance(endpoint_url, str):
            if not s3func.utils.is_url(endpoint_url):
                raise TypeError(f'{endpoint_url} is not a proper url.')
        read_session = s3func.S3Session(access_key_id, access_key, bucket, endpoint_url, max_pool_connections=threads, read_timeout=read_timeout, stream=False, max_attempts=retries)
        key = db_key
    else:
        read_session = None

    return read_session, key


def create_s3_write_session(
        access_key_id: str=None,
        access_key: str=None,
        db_key: str=None,
        bucket: str=None,
        endpoint_url: str=None,
        threads: int=20,
        read_timeout: int=60,
        retries: int=3,
        ):
    """

    """
    if check_write_config(access_key_id, access_key, db_key, bucket, endpoint_url):
        write_session = s3func.S3Session(access_key_id, access_key, bucket, endpoint_url, max_pool_connections=threads, read_timeout=read_timeout, stream=False, max_attempts=retries)
    else:
        write_session = None

    return write_session, db_key









###############################################
### Classes

class JsonSerializer:
    def to_dict(self):
        d1 = dict(db_key=self.db_key,
                  bucket=self.bucket,
                  endpoint_url=self.endpoint_url,
                  db_url=self.db_url,
                  )

        return d1

    def dumps(self):
        return orjson.dumps(self.to_dict())


class S3SessionReader:
    """

    """
    def __init__(self,
                 read_session,
                 read_db_key,
                 threads,
                 ):
        self._read_session = read_session
        self.read_db_key = read_db_key
        self.threads = threads
        self._load_db_metadata()

    def __bool__(self):
        """
        Test to see if remote read access is possible. Same as .read_access.
        """
        return self.read_access

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        """
        Close the remote connection. Should return None.
        """
        if hasattr(self, '_finalizer'):
            self._finalizer()

    def _load_db_metadata(self):
        """
        Load the db metadata from the remote.
        """
        resp_obj = self.head_object()
        if resp_obj.status == 200:
            meta = resp_obj.metadata
            self._init_bytes = base64.urlsafe_b64decode(meta['init_bytes'])
            self.timestamp = int(meta['timestamp'])
            self.uuid = uuid.UUID(hex=meta['uuid'])
            self.type = meta['type']
            self.num_groups = int(meta['num_groups']) if 'num_groups' in meta else None
        elif resp_obj.status == 404:
            self._init_bytes = None
            self.uuid = None
            self.timestamp = None
            self.type = None
            self.num_groups = None
        else:
            raise urllib3.exceptions.HTTPError(resp_obj.error)


    def get_uuid(self):
        """
        Get the UUID of the remote.
        """
        if self.uuid is None:
            self._load_db_metadata()

        return self.uuid


    def get_type(self):
        """
        Get the EBooklet type of the remote.
        """
        if self.type is None:
            self._load_db_metadata()

        return self.type


    def get_timestamp(self):
        """
        Get the metadata timestamp of the remote.
        """
        resp_obj = self.head_object()
        if resp_obj.status == 200:
            self.timestamp = int(resp_obj.metadata['timestamp'])

        elif resp_obj.status == 404:
            self.timestamp = None
        else:
            raise urllib3.exceptions.HTTPError(resp_obj.error)

        return self.timestamp

    def get_user_metadata(self):
        """
        Get the user metadata.
        """
        s3_key = '_metadata' if self.num_groups is not None else f'{booklet.utils.metadata_key_bytes.decode()}'
        resp_obj = self.get_object(s3_key)
        if resp_obj.status == 200:
            meta = orjson.loads(resp_obj.data)
        elif resp_obj.status == 404:
            meta = None
        else:
            raise urllib3.exceptions.HTTPError(resp_obj.error)

        return meta


    def get_object(self, key: str=None, range_start: int=None, range_end: int=None):
        """
        Get a remote object/file. The input should be a key as a str. It should return an object with a .status attribute as an int, a .data attribute in bytes, and a .error attribute as a dict.
        Optionally specify range_start and range_end for byte-range requests.
        """
        if key is None:
            resp = self._read_session.get_object(self.read_db_key, range_start=range_start, range_end=range_end)
        else:
            resp = self._read_session.get_object(self.read_db_key + '/' + key, range_start=range_start, range_end=range_end)

        return resp

    def head_object(self, key: str=None):
        """
        Get the header for a remote object/file. The input should be a key as a str. It should return an object with a .status attribute as an int, a .data attribute in bytes, and a .error attribute as a dict.
        """
        if key is None:
            resp = self._read_session.head_object(self.read_db_key)
        else:
            resp = self._read_session.head_object(self.read_db_key + '/' + key)

        return resp


class S3SessionWriter(S3SessionReader):
    """

    """
    def __init__(self,
                 read_session,
                 write_session,
                 read_db_key,
                 write_db_key,
                 threads,
                 ):
        self._read_session = read_session
        self._write_session = write_session
        self.read_db_key = read_db_key
        self.write_db_key = write_db_key
        self.threads = threads

        self._writable_check = False
        self._writable = False

        ## Finalizer
        self._finalizer = weakref.finalize(self, utils.s3session_finalizer, self._write_session)

        ## Get latest metadata
        self._load_db_metadata()

    @property
    def writable(self):
        """
        Check to see if the remote is writable given the credentials.

        Should I include this? Or should I simply let the other methods fail if it's not writable? I do like having an explicit test...
        """
        if not self._writable_check:
            test_key = self.write_db_key + uuid.uuid6().hex[-13:]
            put_resp = self._write_session.put_object(test_key, b'0')
            if put_resp.status // 100 == 2:
                del_resp = self._write_session.delete_object(test_key, put_resp.metadata['version_id'])
                if del_resp.status // 100 == 2:
                    self._writable = True

            self._writable_check = True

        return self._writable


    def put_db_object(self, data: bytes, metadata):
        """
        Upload the main db object to the remote.
        """
        if self.writable:
            return self._write_session.put_object(self.write_db_key, data, metadata=metadata)
        else:
            raise ValueError('Session is not writable.')


    def put_object(self, key: str, data: bytes, metadata=None):
        """
        Upload an object to the remote.
        """
        if metadata is None:
            metadata = {}
        if self.writable:
            key1 = self.write_db_key + '/' + key
            return self._write_session.put_object(key1, data, metadata=metadata)
        else:
            raise ValueError('Session is not writable.')


    def delete_object(self, key: str):
        """
        Delete a single object.
        """
        if self.writable:
            key1 = self.write_db_key + '/' + key
            resp = self._write_session.list_object_versions(prefix=key1)
            del_list = []
            for obj in resp.iter_objects():
                del_list.append({'key': obj['key'], 'version_id': obj['version_id']})
            if del_list:
                return self._write_session.delete_objects(del_list)
        else:
            raise ValueError('Session is not writable.')

    def delete_objects(self, keys):
        """
        Delete specific objects.
        """
        if self.writable:
            del_list = []
            resp = self._write_session.list_object_versions(prefix=self.write_db_key + '/')
            for obj in resp.iter_objects():
                key0 = obj['key']
                key = key0.split('/')[-1]
                if key in keys:
                    del_list.append({'key': key0, 'version_id': obj['version_id']})

            if del_list:
                del_resp = self._write_session.delete_objects(del_list)
                if del_resp.status // 100 != 2:
                    raise urllib3.exceptions.HTTPError(del_resp.error)
        else:
            raise ValueError('Session is not writable.')


    def delete_remote(self):
        """
        Delete the entire remote.
        """
        if self.writable:
            del_list = []
            resp = self._write_session.list_object_versions(prefix=self.write_db_key)
            for obj in resp.iter_objects():
                key0 = obj['key']
                del_list.append({'key': key0, 'version_id': obj['version_id']})

            self._write_session.delete_objects(del_list)
            self._init_bytes = None
            self.uuid = None
        else:
            raise ValueError('Session is not writable.')

    def copy_remote(self, remote_conn):
        """
        Copy an entire remote dataset to another remote location. The new location must be empty.

        There's still a question of whether this method should be in the Reader rather than the writer. As a matter of API concept, this should be in the Reader as you only need to read something to copy it somewhere else (the target). But as a matter of implementation, the thing doing the copying would need to know all of the files to copy. This can be either known through a list_objects call (currently used and requires write permissions as defined in this API) or to parse the index file for all of the appropriate keys. Parsing the index file requires the file to be saved to disk and this API does not have that capability. Consequently, if the index file must be used, then it must be implemented in the EVariableLengthValue class instead of the session classes. Using the index file instead of the list_objects method would have the advantage of only copying over the objects that are actually used by ebooklet rather than other dangling objects. On the other hand, requiring the source session being a Writer would more likely cause the copy to be an effecient S3 to S3 copy rather than a slower routing through the user's network.
        """
        with remote_conn.open('w') as writer:

            if not writer.writable:
                raise ValueError('target remote is not writable.')

            ## Check if source exists
            source_uuid = self.get_uuid()
            if source_uuid is None:
                raise ValueError('The source remote does not exist.')

            ## Check is target exists
            target_uuid = writer.get_uuid()
            if target_uuid is not None:
                raise ValueError('The target remote already exists. Either delete_remote or use a different target.')

            ## Get a list of all source objects
            source_resp = self._write_session.list_objects(prefix=self.write_db_key + '/')
            if source_resp.status // 100 != 2:
                raise urllib3.exceptions.HTTPError(source_resp.error)

            ## Get all target objects
            target_resp = writer._write_session.list_objects(prefix=writer.write_db_key + '/')
            if target_resp.status // 100 != 2:
                raise urllib3.exceptions.HTTPError(target_resp.error)

            target_exist_keys = set(obj['key'] for obj in target_resp.iter_objects())

            ## Buckets
            source_bucket = self._write_session.bucket
            target_bucket = writer._write_session.bucket

            ## Determine if the s3 copy_object method can be used
            if (self._write_session._access_key_id == writer._write_session._access_key_id) and (self._write_session._access_key == writer._write_session._access_key):
                logger.info('Both the source and target remotes use the same credentials, so copying objects is efficient.')

                futures = {}
                failures = {}
                with concurrent.futures.ThreadPoolExecutor(max_workers=self.threads) as executor:
                    for obj in source_resp.iter_objects():
                        source_key = obj['key']
                        target_key = writer.write_db_key + source_key.removeprefix(self.write_db_key)
                        if target_key not in target_exist_keys:
                            f = executor.submit(self._write_session.copy_object, source_key, target_key, source_bucket=source_bucket, dest_bucket=target_bucket)
                            futures[f] = target_key

                    for f in concurrent.futures.as_completed(futures):
                        target_key = futures[f]
                        resp = f.result()
                        if resp.status // 100 != 2:
                            failures[target_key] = resp.error

                if failures:
                    logger.warning('Copy failures have occurred. Rerun copy_remote or delete_remote.')
                    return failures
                else:
                    resp = self._write_session.copy_object(self.write_db_key, writer.write_db_key, source_bucket=source_bucket, dest_bucket=target_bucket)
                    if resp.status // 100 != 2:
                        raise urllib3.exceptions.HTTPError(resp.error)

            else:
                logger.info('The source and target remotes use different credentials, so copying objects must be first downloaded then uploaded. Less efficient than if both remotes had the same credentials.')

                futures = {}
                failures = {}
                with concurrent.futures.ThreadPoolExecutor(max_workers=self.threads) as executor:
                    for obj in source_resp.iter_objects():
                        source_key = obj['key']
                        target_key = writer.write_db_key + source_key.removeprefix(self.write_db_key)
                        if target_key not in target_exist_keys:
                            f = executor.submit(utils.indirect_copy_remote, self._read_session, writer._write_session, source_key, target_key, source_bucket=source_bucket, dest_bucket=target_bucket)
                            futures[f] = target_key

                    for f in concurrent.futures.as_completed(futures):
                        target_key = futures[f]
                        resp = f.result()
                        if resp.status // 100 != 2:
                            failures[target_key] = resp.error

                if failures:
                    logger.warning('Copy failures have occurred. Rerun copy_remote or delete_remote.')
                    return failures
                else:
                    resp = utils.indirect_copy_remote(self._read_session, writer._write_session, self.write_db_key, writer.write_db_key, source_bucket, target_bucket)
                    if resp.status // 100 != 2:
                        raise urllib3.exceptions.HTTPError(resp.error)


    def list_objects(self):
        """

        """
        return self._write_session.list_objects(prefix=self.write_db_key + '/')


    def list_object_versions(self):
        """

        """
        return self._write_session.list_object_versions(prefix=self.write_db_key + '/')

    def create_lock(self):
        """
        Initialise an S3 lock object. A lock is not immediately aquired. This must be done via the lock object (as well as releasing locks).
        """
        if self.writable:
            lock = self._write_session.lock(self.write_db_key)
            return lock
        else:
            raise ValueError('Session is not writable.')

    def break_other_locks(self, timestamp: str | datetime.datetime=None):
        """
        Removes all locks that are on the object older than specified timestamp. This is only meant to be used in deadlock circumstances.

        Parameters
        ----------
        timestamp : str or datetime.datetime
            All locks older than the timestamp will be removed. The default is now.

        Returns
        -------
        list of dict of the removed keys/versions
        """
        if self.writable:
            lock = self._write_session.lock(self.write_db_key)
            other_keys = lock.break_other_locks(timestamp)

            return other_keys
        else:
            raise ValueError('Session is not writable.')


    def _head_object_writer(self, key: str=None):
        """
        Get the header for a remote object/file. The input should be a key as a str. It should return an object with a .status attribute as an int, a .data attribute in bytes, and a .error attribute as a dict.
        """
        if key is None:
            resp = self._write_session.head_object(self.write_db_key)
        else:
            resp = self._write_session.head_object(self.write_db_key + '/' + key)

        return resp

    def _get_uuid_writer(self):
        """
        Get the uuid of the remote file.
        """
        resp_obj = self._head_object_writer()
        if resp_obj.status == 200:
            uuid1 = uuid.UUID(hex=resp_obj.metadata['uuid'])
        elif resp_obj.status == 404:
            uuid1 = None
        else:
            raise urllib3.exceptions.HTTPError(resp_obj.error)

        return uuid1


class S3Connection(JsonSerializer):
    """

    """
    def __init__(self,
                access_key_id: str=None,
                access_key: str=None,
                db_key: str=None,
                bucket: str=None,
                endpoint_url: str=None,
                db_url: str=None,
                threads: int=20,
                read_timeout: int=60,
                retries: int=3,
                ):
        """
        Establishes an S3 client connection with an S3 account.

        Parameters
        ----------
        access_key_id : str
            The access key id also known as aws_access_key_id.
        access_key : str
            The access key also known as aws_secret_access_key.
        db_key : str
            The key name of the database.
        bucket : str
            The bucket to be used when performing S3 operations.
        endpoint_url : str
            The nedpoint http(s) url for the s3 service.
        threads : int
            The number of simultaneous connections for the S3 connection.
        read_timeout: int
            The read timeout in seconds passed to the "retries" option in the S3 config.
        retries: int
            The number of max attempts passed to the "retries" option in the S3 config.
        """
        ## temp read session
        read_session, key = create_s3_read_session(
                access_key_id,
                access_key,
                db_key,
                bucket,
                endpoint_url,
                db_url,
                threads,
                read_timeout,
                retries,
                )
        if read_session is None:
            raise ValueError('Either db_url or a combo of access_key_id, access_key, db_key, and bucket (and optionally endpoint_url) must be passed.')

        ## Assign properties
        self.db_key = db_key
        self.bucket = bucket
        self.access_key_id = access_key_id
        self.access_key = access_key
        self.endpoint_url = endpoint_url
        self.db_url = db_url
        self.threads = threads
        self.read_timeout = read_timeout
        self.retries = retries

    def open(self,
             flag: str='r',
             ):
        """
        Opens a connection to the S3 remote.

        Parameters
        ----------
        flag : str
            The open flag for the remote. These are the same for booklet and ebooklet.

        Returns
        -------
        S3SessionReader or S3SessionWriter
        """
        if (flag != 'r') and (self.access_key_id is None or self.access_key is None):
            raise ValueError("access_key_id and access_key must be assigned to open a connection for writing.")

        ## Read session
        read_session, read_db_key = create_s3_read_session(
                self.access_key_id,
                self.access_key,
                self.db_key,
                self.bucket,
                self.endpoint_url,
                self.db_url,
                self.threads,
                self.read_timeout,
                self.retries,
                )

        if flag == 'r':
            return S3SessionReader(read_session, read_db_key, self.threads)
        else:
            write_session, write_db_key = create_s3_write_session(
                    self.access_key_id,
                    self.access_key,
                    self.db_key,
                    self.bucket,
                    self.endpoint_url,
                    self.threads,
                    self.read_timeout,
                    self.retries,
                    )

            if write_session is None:
                raise ValueError("A write session could not be created. Check that all the inputs are assigned.")

            session_writer = S3SessionWriter(
                                    read_session,
                                    write_session,
                                    read_db_key,
                                    write_db_key,
                                    self.threads,
                                    )

            # Check to make sure the uuids are the same if the read and write sessions are different
            if isinstance(read_session, s3func.HttpSession) and session_writer.uuid is not None:
                if session_writer._get_uuid_writer() != session_writer.uuid:
                    raise ValueError('The UUIDs of the http connection and the S3 connection are different. Check to make sure the they are pointing to the right file.')

            return session_writer



















































#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""
from collections.abc import Mapping, MutableMapping
from typing import Union
import pathlib
from concurrent.futures import ThreadPoolExecutor, as_completed
import booklet
import weakref
from itertools import count
from collections import deque
import urllib3

from . import utils
from . import remote


#######################################################
### Classes


_MISSING = object()


class Change:
    """

    """
    def __init__(self, ebooklet):
        """
        Open a Change object to interact with the remote.
        """
        ebooklet.sync()

        self._ebooklet = ebooklet

        self._changelog_path = None


    def pull(self):
        """
        Update the remote index file to determine if any changes have occurred.
        """
        self._ebooklet.sync()

        ## update the remote timestamp
        self._ebooklet._remote_session.get_timestamp()

        ## Determine if a change has occurred
        overwrite_remote_index = utils.check_local_remote_sync(self._ebooklet._local_file, self._ebooklet._remote_session, self._ebooklet._flag)

        ## Pull down the remote index
        if overwrite_remote_index:
            utils.get_remote_index_file(self._ebooklet._local_file_path, overwrite_remote_index, self._ebooklet._remote_session, self._ebooklet._flag)


    def update(self):
        """
        Determine if there are any changes between the local and remote databases.
        """
        self._ebooklet.sync()
        changelog_path = utils.create_changelog(self._ebooklet._local_file_path, self._ebooklet._local_file, self._ebooklet._remote_index, self._ebooklet._remote_session)

        self._changelog_path = changelog_path


    def iter_changes(self):
        """
        Create an iterator of the changes.
        """
        if not self._changelog_path:
            self.update()
        return utils.view_changelog(self._changelog_path)


    def discard(self, keys=None):
        """
        Removes changed keys in the local file. If keys is None, then removes all changed keys.
        """
        if not self._ebooklet.writable:
            raise ValueError('File is open for read-only.')

        if not self._changelog_path:
            self.update()

        with booklet.FixedLengthValue(self._changelog_path) as f:
            if keys is None:
                rm_keys = f.keys()
            else:
                rm_keys = [key for key in keys if key in f]

            for key in rm_keys:
                # print(key)
                del self._ebooklet._local_file[key]

        self._changelog_path.unlink()
        self._changelog_path = None


    def push(self, force_push=False):
        """
        Updates the remote. It will regenerate the changelog to ensure the changelog is up-to-date. Returns True if the remote has been updated and False if no updates were made (due to nothing needing updating). If upload failures have occurred, then it will return a list of the keys that failed.
        Force_push will push the main file and the remote_index to the remote regardless of changes. Only necessary if upload failures occurred during a previous push.
        """
        if not self._ebooklet._remote_session.writable:
            raise ValueError('Remote is not writable.')

        if not self._ebooklet.writable:
            raise ValueError('File is open for read-only.')

        self.update()

        result = utils.update_remote(self._ebooklet._local_file, self._ebooklet._remote_index, self._changelog_path, self._ebooklet._remote_session, force_push, self._ebooklet._deletes, self._ebooklet._flag, self._ebooklet.type, self._ebooklet._num_groups)

        if isinstance(result, dict):
            # Partial failure — don't clean up changelog so push can be retried
            return result

        if result:
            self._changelog_path.unlink()
            self._changelog_path = None
            self._ebooklet._deletes.clear()

            if self._ebooklet._remote_session.uuid is None:
                self._ebooklet._remote_session._load_db_metadata()

        return result


class EVariableLengthValue(MutableMapping):
    """

    """
    def __init__(
            self,
            remote_session: remote.S3SessionReader | remote.S3SessionWriter,
            local_file_path: pathlib.Path,
            flag: str = "r",
            value_serializer: str = None,
            n_buckets: int=12007,
            buffer_size: int = 2**22,
            num_groups: int = None,
            ):
        """

        """
        ## Remove the remote database if flag == 'n'
        if flag == 'n' and remote_session.uuid is not None:
            remote_session.delete_remote()

        self._init_common(remote_session, local_file_path, flag, value_serializer, n_buckets, buffer_size, 'EVariableLengthValue', num_groups)

    def _init_common(self, remote_session, local_file_path, flag, value_serializer, n_buckets, buffer_size, ebooklet_type, num_groups=None):
        """
        Shared initialization logic for EVariableLengthValue and RemoteConnGroup.
        """
        ## Lock the remote if file is opened for write
        if flag != 'r':
            lock = remote_session.create_lock()
            lock.acquire()
            self.writable = True
        else:
            lock = None
            self.writable = False

        ## Init the local file
        local_file, overwrite_remote_index = utils.init_local_file(local_file_path, flag, remote_session, value_serializer, n_buckets, buffer_size)

        remote_index_path = utils.get_remote_index_file(local_file_path, overwrite_remote_index, remote_session, flag)

        ## Open remote index file
        if remote_index_path.exists():
            if flag == 'r':
                remote_index = booklet.FixedLengthValue(remote_index_path, 'r')
            else:
                remote_index = booklet.FixedLengthValue(remote_index_path, 'w')
        else:
            remote_index = booklet.FixedLengthValue(remote_index_path, 'n', key_serializer='str', value_len=15, n_buckets=n_buckets, buffer_size=buffer_size)

        ## Resolve num_groups: remote metadata > user param
        if remote_session.num_groups is not None:
            resolved_num_groups = remote_session.num_groups
        else:
            resolved_num_groups = num_groups

        ## Finalizer
        self._finalizer = weakref.finalize(self, utils.ebooklet_finalizer, local_file, remote_index, remote_session, lock)

        ## Assign properties
        self._flag = flag
        self.lock = lock
        self._local_file_path = local_file_path
        self._local_file = local_file
        self._remote_index_path = remote_index_path
        self._remote_index = remote_index
        self._deletes = set()
        self._remote_session = remote_session
        self._n_buckets = local_file._n_buckets
        self.type = ebooklet_type
        self._num_groups = resolved_num_groups


    def set_metadata(self, data, timestamp=None):
        """
        Sets the metadata for the booklet. The data input must be a json serializable object. Optionally assign a timestamp.
        """
        if self.writable:
            self._local_file.set_metadata(data, timestamp)
        else:
            raise ValueError('File is open for read only.')


    def get_metadata(self, include_timestamp=False):
        """
        Get the metadata. Optionally include the timestamp in the output.
        Will return None if no metadata has been assigned.
        """
        failure = self._load_item(booklet.utils.metadata_key_bytes.decode())
        if failure:
            raise urllib3.exceptions.HTTPError(failure)

        self._local_file.sync()

        return self._local_file.get_metadata(include_timestamp=include_timestamp)


    def keys(self):
        """
        Returns a generator of the keys.
        """
        overlap = {utils.metadata_key_str}
        for key in self._local_file.keys():
            if key in self._remote_index:
                overlap.add(key)
            yield key

        for key in self._remote_index.keys():
            if key not in overlap:
                yield key


    def items(self):
        """
        Returns an iterator of the keys, values.
        """
        failure_dict = self.load_items()

        if failure_dict:
            raise urllib3.exceptions.HTTPError(failure_dict)

        return self._local_file.items()


    def values(self):
        """
        Returns an iterator of the values.
        """
        failure_dict = self.load_items()

        if failure_dict:
            raise urllib3.exceptions.HTTPError(failure_dict)

        return self._local_file.values()

    def timestamps(self, include_value=False):
        """
        Return an iterator for timestamps for all keys. Optionally add values to the iterator.
        """
        failure_dict = self.load_items()

        if failure_dict:
            raise urllib3.exceptions.HTTPError(failure_dict)

        return self._local_file.timestamps(include_value=include_value)


    def get_timestamp(self, key, include_value=False, decode_value=True, default=None):
        """
        Get a timestamp associated with a key. Optionally include the value.
        """
        failure = self._load_item(key)
        if failure:
            raise urllib3.exceptions.HTTPError(failure)

        return self._local_file.get_timestamp(key, include_value=include_value, decode_value=decode_value, default=default)

    def set_timestamp(self, key, timestamp):
        """
        Set a timestamp for a specific key. The timestamp must be either an int of the number of microseconds in POSIX UTC time, an ISO 8601 datetime string with timezone, or a datetime object with timezone.
        """
        if self.writable:
            self._local_file.set_timestamp(key, timestamp)
        else:
            raise ValueError('File is open for read only.')


    def set(self, key, value, timestamp=None, encode_value=True):
        """
        Set a value associated with a key.
        """
        if self.writable:
            self._local_file.set(key, value, timestamp=timestamp, encode_value=encode_value)

        else:
            raise ValueError('File is open for read only.')


    def __iter__(self):
        return self.keys()

    def __len__(self):
        """

        """
        counter = count()
        deque(zip(self.keys(), counter), maxlen=0)

        return next(counter)


    def __contains__(self, key):
        if (key in self._remote_index) or (key in self._local_file):
            return True
        else:
            return False

    def get(self, key, default=None):
        """
        Get a value associated with a key. Will return the default if the key does not exist.
        """
        failure = self._load_item(key)
        if failure:
            raise urllib3.exceptions.HTTPError(failure)

        return self._local_file.get(key, default=default)


    def update(self, key_value: MutableMapping):
        """
        Set many keys/values from a dict.
        """
        if self.writable:
            for key, value in key_value.items():
                self[key] = value
        else:
            raise ValueError('File is open for read only.')


    def prune(self, timestamp=None):
        """
        Prunes the old keys and associated values. Returns the number of removed items. The method can also prune remove keys/values older than the timestamp. The user can also reindex the booklet file. False does no reindexing, True increases the n_buckets to a preassigned value, or an int of the n_buckets. True can only be used if the default n_buckets were used at original initialisation.
        """
        if self.writable:
            removed = self._local_file.prune(timestamp=timestamp)
            self._n_buckets = self._local_file._n_buckets

            _ = self._remote_index.prune()

            return removed
        else:
            raise ValueError('File is open for read only.')


    def get_items(self, keys, default=None):
        """
        Return an iterator of the values associated with the input keys. It will first load all of the keys/values into the local file. Missing keys will return the default.
        """
        if not isinstance(keys, (list, tuple, set)):
            keys = tuple(keys)

        failure_dict = self.load_items(keys)

        if failure_dict:
            raise urllib3.exceptions.HTTPError(failure_dict)

        for key in keys:
            output = self._local_file.get(key, default=default)
            yield key, output


    def load_items(self, keys=None):
        """
        Loads items into the local file from the remote. If keys is None, then it loads all of the values from the remote in to the local file. Returns a dict of failed transfers.
        """
        futures = {}
        failure_dict = {}

        with ThreadPoolExecutor(max_workers=self._remote_session.threads) as executor:
            if keys is None:
                items_iter = self._remote_index.items()
            else:
                items_iter = ((k, self._remote_index.get(k)) for k in keys)

            if self._num_groups is not None:
                groups_to_download = {}
                for key, remote_val in items_iter:
                    remote_time_bytes = remote_val[:7] if remote_val else None
                    check = utils.check_local_vs_remote(self._local_file, remote_time_bytes, key)
                    if check:
                        if key == utils.metadata_key_str:
                            f = executor.submit(utils.get_remote_value, self._local_file, key, self._remote_session)
                            futures[f] = key
                        else:
                            group_id = utils.key_to_group_id(key, self._num_groups)
                            offset = utils.bytes_to_int(remote_val[7:11])
                            length = utils.bytes_to_int(remote_val[11:15])
                            timestamp_int = utils.bytes_to_int(remote_val[:7])
                            groups_to_download.setdefault(group_id, []).append((key, offset, length, timestamp_int))

                for group_id, key_infos in groups_to_download.items():
                    f = executor.submit(utils.get_remote_group_values, group_id, key_infos, self._local_file, self._remote_session)
                    futures[f] = f'_group_{group_id}'
            else:
                for key, remote_val in items_iter:
                    remote_time_bytes = remote_val[:7] if remote_val else None
                    check = utils.check_local_vs_remote(self._local_file, remote_time_bytes, key)
                    if check:
                        f = executor.submit(utils.get_remote_value, self._local_file, key, self._remote_session)
                        futures[f] = key

            for f in as_completed(futures):
                key = futures[f]
                error = f.result()
                if error is not None:
                    failure_dict[key] = error

        return failure_dict


    def _load_item(self, key):
        """

        """
        remote_val = self._remote_index.get(key)
        remote_time_bytes = remote_val[:7] if remote_val else None
        check = utils.check_local_vs_remote(self._local_file, remote_time_bytes, key)

        if check:
            if self._num_groups is not None and key != utils.metadata_key_str:
                group_id = utils.key_to_group_id(key, self._num_groups)
                offset = utils.bytes_to_int(remote_val[7:11])
                length = utils.bytes_to_int(remote_val[11:15])
                timestamp_int = utils.bytes_to_int(remote_val[:7])
                failure = utils.get_remote_group_value(group_id, key, offset, length, timestamp_int, self._local_file, self._remote_session)
            else:
                failure = utils.get_remote_value(self._local_file, key, self._remote_session)
            return failure
        else:
            return None


    def __getitem__(self, key: str):
        value = self.get(key, default=_MISSING)

        if value is _MISSING:
            raise KeyError(f'{key}')
        else:
            return value


    def __setitem__(self, key: str, value):
        if self.writable:
            self.set(key, value)
        else:
            raise ValueError('File is open for read only.')

    def __delitem__(self, key):
        if self.writable:
            if key in self._remote_index:
                del self._remote_index[key]
                self._deletes.add(key)

            if key in self._local_file:
                del self._local_file[key]
        else:
            raise ValueError('File is open for read only.')

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def clear(self):
        """
        Remove all keys and values from the local file.
        """
        if self.writable:
            self._local_file.clear()

        else:
            raise ValueError('File is open for read only.')

    def close(self, force_shutdown=False):
        """
        Close all open objects. If force_shutdown is True, then it will immediately end any processes running in the background (not recommended unless there's a deadlock).
        """
        self.sync()
        self._finalizer()


    def sync(self, force_shutdown=False):
        """
        Syncronize all cache to disk. This ensures all data has been saved to disk properly. If force_shutdown is True, then it will immediately end any processes running in the background (not recommended unless there's a deadlock).
        """
        self._remote_index.sync()
        self._local_file.sync()

    def changes(self):
        """
        Return a Change object of the changes that have occurred during this session.
        """
        return Change(self)


    def delete_remote(self):
        """
        Completely delete the remote file, but keep the local file.
        """
        if self.writable:
            self._remote_session.delete_remote()
        else:
            raise ValueError('File is open for read only.')


    def copy_remote(self, remote_conn):
        """
        Copy the entire remote file to another remote location. The new location must be empty.
        """
        if self.writable:
            self._remote_session.copy_remote(remote_conn)
        else:
            raise ValueError('File is open for read only.')

    def map(self, func, keys=None, n_workers=None):
        """
        Apply func to items in parallel using multiprocessing.

        Parameters
        ----------
        func : callable
            A picklable function: func(key, value) -> (new_key, new_value) or None.
            Return a (key, value) tuple to yield the result. The output key can
            differ from the input key. Return None to skip (item not yielded).
            Must be a top-level function (not a lambda or closure).
        keys : iterable, optional
            Specific keys to process. If None, processes all keys.
        n_workers : int, optional
            Number of worker processes. Defaults to os.cpu_count().

        Yields
        ------
        tuple
            (key, value) pairs produced by func, as they complete.
        """
        if keys is not None and not isinstance(keys, (list, tuple)):
            keys = list(keys)

        failure_dict = self.load_items(keys)
        if failure_dict:
            raise urllib3.exceptions.HTTPError(failure_dict)

        yield from self._local_file.map(func, keys=keys, n_workers=n_workers)


class RemoteConnGroup(EVariableLengthValue):
    """

    """
    def __init__(
            self,
            remote_session: remote.S3SessionReader | remote.S3SessionWriter,
            local_file_path: pathlib.Path,
            flag: str = "r",
            n_buckets: int=12007,
            buffer_size: int = 2**22,
            num_groups: int = None,
            ):
        """

        """
        self._init_common(remote_session, local_file_path, flag, 'orjson', n_buckets, buffer_size, 'RemoteConnGroup', num_groups)


    def add(self, remote_conn: remote.S3Connection):
        """
        Add a remote connection to the group.
        """
        if self.writable:


            if not isinstance(remote_conn, remote.S3Connection):
                raise TypeError('remote_conn/value must be a remote.S3Connection')

            ## Get remote_conn metadata
            with remote_conn.open() as rc:
                uuid0 = rc.get_uuid()
                if uuid0 is None:
                    raise ValueError('Remote does not exist. It must exist to be added to a RemoteConnGroup.')

                uuid_hex = uuid0.hex

                user_meta = rc.get_user_metadata()
                ebooklet_type = rc.get_type()
                ts = rc.get_timestamp()

            conn_dict = {'type': ebooklet_type,
                         'timestamp': ts,
                         'user_meta': user_meta,
                         'remote_conn': remote_conn.to_dict(),
                         }

            self._local_file.set(uuid_hex, conn_dict, ts)

        else:
            raise ValueError('File is open for read only.')


    def set(self, key, remote_conn: remote.S3Connection):
        """
        Use the add method to add remote connections to the group.
        """
        raise NotImplementedError('Use the add method to add remote connections to the group.')


def open_ebooklet(
    remote_conn: Union[remote.S3Connection, str, dict],
    file_path: Union[str, pathlib.Path],
    flag: str = "r",
    value_serializer: str = None,
    n_buckets: int=12007,
    buffer_size: int = 2**22,
    num_groups: int = None,
    ):
    """
    Open an S3 dbm-style database. This allows the user to interact with an S3 bucket like a MutableMapping (python dict) object.

    Parameters
    -----------
    remote_conn : S3Connection, str, or dict
        The object to connect to a remote. It can be an S3Connection object, an http url string, or a dict with the parameters for initializing an S3Connection object.

    file_path : str or pathlib.Path
        It must be a path to a local file location. If you want to use a tempfile, then use the name from the NamedTemporaryFile initialized class.

    flag : str
        Flag associated with how the file is opened according to the dbm style. See below for details.

    value_serializer : str, class, or None
        The serializer to use to convert the input value to bytes. Run the booklet.available_serializers to determine the internal serializers that are available. None will require bytes as input.

    n_buckets : int
        The number of hash buckets to using in the indexing. Generally use the same number of buckets as you expect for the total number of keys.

    buffer_size : int
        The buffer memory size in bytes used for writing. Writes are first written to a block of memory, then once the buffer if filled up it writes to disk. This is to reduce the number of writes to disk and consequently the CPU write overhead.
        This is only used when the file is open for writing.

    num_groups : int or None
        The number of groups for grouped S3 object storage. Required when creating a new database (flag='n'). For existing databases, this value is read from S3 metadata and the user-provided value is ignored. If None for a new database, per-key storage is used.

    Returns
    -------
    EVariableLengthValue

    The optional *flag* argument can be:

    +---------+-------------------------------------------+
    | Value   | Meaning                                   |
    +=========+===========================================+
    | ``'r'`` | Open existing database for reading only   |
    |         | (default)                                 |
    +---------+-------------------------------------------+
    | ``'w'`` | Open existing database for reading and    |
    |         | writing                                   |
    +---------+-------------------------------------------+
    | ``'c'`` | Open database for reading and writing,    |
    |         | creating it if it doesn't exist           |
    +---------+-------------------------------------------+
    | ``'n'`` | Always create a new, empty database, open |
    |         | for reading and writing                   |
    +---------+-------------------------------------------+
    """
    if num_groups is not None and num_groups < 1:
        raise ValueError('num_groups must be a positive integer.')

    local_file_path = pathlib.Path(file_path)

    local_file_exists = local_file_path.exists()

    ## Check and open the remote session
    remote_conn = remote.check_remote_conn(remote_conn, flag)
    remote_session, ebooklet_type = utils.open_remote_conn(remote_conn, flag, local_file_exists)

    if ebooklet_type is not None and ebooklet_type != 'EVariableLengthValue':
        raise TypeError(f'The remote database is of type {ebooklet_type}, not EVariableLengthValue. Use open_rcg() instead.')

    return EVariableLengthValue(remote_session=remote_session, local_file_path=local_file_path, flag=flag, value_serializer=value_serializer, n_buckets=n_buckets, buffer_size=buffer_size, num_groups=num_groups)


def open_rcg(
    remote_conn: Union[remote.S3Connection, str, dict],
    file_path: Union[str, pathlib.Path],
    flag: str = "r",
    n_buckets: int=12007,
    buffer_size: int = 2**22,
    num_groups: int = None,
    ):
    """
    Open an S3-backed remote connection group. A remote connection group stores S3Connection references as key-value pairs, using orjson serialization.

    Parameters
    -----------
    remote_conn : S3Connection, str, or dict
        The object to connect to a remote. It can be an S3Connection object, an http url string, or a dict with the parameters for initializing an S3Connection object.

    file_path : str or pathlib.Path
        It must be a path to a local file location. If you want to use a tempfile, then use the name from the NamedTemporaryFile initialized class.

    flag : str
        Flag associated with how the file is opened according to the dbm style. See below for details.

    n_buckets : int
        The number of hash buckets to using in the indexing. Generally use the same number of buckets as you expect for the total number of keys.

    buffer_size : int
        The buffer memory size in bytes used for writing. Writes are first written to a block of memory, then once the buffer if filled up it writes to disk. This is to reduce the number of writes to disk and consequently the CPU write overhead.
        This is only used when the file is open for writing.

    num_groups : int or None
        The number of groups for grouped S3 object storage. Required when creating a new database (flag='n'). For existing databases, this value is read from S3 metadata and the user-provided value is ignored. If None for a new database, per-key storage is used.

    Returns
    -------
    RemoteConnGroup

    The optional *flag* argument can be:

    +---------+-------------------------------------------+
    | Value   | Meaning                                   |
    +=========+===========================================+
    | ``'r'`` | Open existing database for reading only   |
    |         | (default)                                 |
    +---------+-------------------------------------------+
    | ``'w'`` | Open existing database for reading and    |
    |         | writing                                   |
    +---------+-------------------------------------------+
    | ``'c'`` | Open database for reading and writing,    |
    |         | creating it if it doesn't exist           |
    +---------+-------------------------------------------+
    | ``'n'`` | Always create a new, empty database, open |
    |         | for reading and writing                   |
    +---------+-------------------------------------------+
    """
    if num_groups is not None and num_groups < 1:
        raise ValueError('num_groups must be a positive integer.')

    local_file_path = pathlib.Path(file_path)

    local_file_exists = local_file_path.exists()

    ## Check and open the remote session
    remote_conn = remote.check_remote_conn(remote_conn, flag)
    remote_session, ebooklet_type = utils.open_remote_conn(remote_conn, flag, local_file_exists)

    if ebooklet_type is not None and ebooklet_type != 'RemoteConnGroup':
        raise TypeError(f'The remote database is of type {ebooklet_type}, not RemoteConnGroup. Use open_ebooklet() instead.')

    return RemoteConnGroup(remote_session=remote_session, local_file_path=local_file_path, flag=flag, n_buckets=n_buckets, buffer_size=buffer_size, num_groups=num_groups)





"""

"""
from ebooklet.main import open_ebooklet, open_rcg, EVariableLengthValue, RemoteConnGroup
from ebooklet.remote import S3Connection

__all__ = ["open_ebooklet", "open_rcg", "EVariableLengthValue", 'RemoteConnGroup', 'S3Connection']

__version__ = '0.8.0'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan  5 11:04:13 2023

@author: mike
"""
import logging
import hashlib
import struct
import booklet
import urllib3
from datetime import datetime, timezone
import base64
import portalocker
from concurrent.futures import ThreadPoolExecutor, as_completed
import orjson

logger = logging.getLogger(__name__)

############################################
### Parameters

int_to_bytes = booklet.utils.int_to_bytes
bytes_to_int = booklet.utils.bytes_to_int
metadata_key_str = booklet.utils.metadata_key_bytes.decode()

############################################
### Group functions


def key_to_group_id(key: str, num_groups: int) -> int:
    digest = hashlib.blake2b(key.encode(), digest_size=4).digest()
    return int.from_bytes(digest, 'big') % num_groups


def pack_group(entries: list[tuple[str, int, bytes]]) -> tuple[bytes, dict[str, tuple[int, int]]]:
    buf = struct.pack('>I', len(entries))
    offsets = {}
    pos = 4
    for key, timestamp, value in entries:
        key_bytes = key.encode()
        buf += struct.pack('>H', len(key_bytes))
        pos += 2
        buf += key_bytes
        pos += len(key_bytes)
        buf += int_to_bytes(timestamp, 7)
        pos += 7
        buf += struct.pack('>I', len(value))
        pos += 4
        offsets[key] = (pos, len(value))
        buf += value
        pos += len(value)
    return buf, offsets


def unpack_group(data: bytes) -> list[tuple[str, int, bytes]]:
    offset = 0
    num_entries = struct.unpack_from('>I', data, offset)[0]
    offset += 4
    entries = []
    for _ in range(num_entries):
        key_len = struct.unpack_from('>H', data, offset)[0]
        offset += 2
        key = data[offset:offset + key_len].decode()
        offset += key_len
        timestamp = bytes_to_int(data[offset:offset + 7])
        offset += 7
        value_len = struct.unpack_from('>I', data, offset)[0]
        offset += 4
        value = data[offset:offset + value_len]
        offset += value_len
        entries.append((key, timestamp, value))
    return entries


############################################
### Functions


def s3session_finalizer(session):
    """
    The finalizer function for S3Remote instances.
    """
    session._session.clear()


def ebooklet_finalizer(local_file, remote_index, remote_session, lock):
    """
    The finalizer function for book instances.
    """
    local_file.close()
    remote_index.close()
    remote_session.close()
    if lock is not None:
        lock.release()


def open_remote_conn(remote_conn, flag, local_file_exists):
    """

    """
    remote_session = remote_conn.open(flag)

    if flag == 'r' and (remote_session.uuid is None) and not local_file_exists:
        raise ValueError('No file was found in the remote, but the local file was open for read without creating a new file.')

    ebooklet_type = remote_session.type

    return remote_session, ebooklet_type


def check_local_remote_sync(local_file, remote_session, flag):
    """

    """
    overwrite_remote_index = False

    remote_uuid = remote_session.uuid
    if remote_uuid and flag != 'n':
        local_uuid = local_file.uuid

        if remote_uuid != local_uuid:
            raise ValueError('The local file has a different UUID than the remote. Use a different local file path or delete the existing one.')

        ## Check timestamp to determine if the local remote_index needs to be updated
        if (remote_session.timestamp > local_file._file_timestamp):
            overwrite_remote_index = True

    return overwrite_remote_index


def init_local_file(local_file_path, flag, remote_session, value_serializer, n_buckets, buffer_size):
    """

    """
    remote_uuid = remote_session.uuid

    if local_file_path.exists():

        if flag == 'n':
            local_file = booklet.open(local_file_path, flag='n', key_serializer='str', value_serializer=value_serializer, n_buckets=n_buckets, buffer_size=buffer_size)

            overwrite_remote_index = True
        else:
            ## The local file will need to always be open for write since data will be loaded from the remote regardless if the user has only opened it for read-only
            local_file = booklet.open(local_file_path, 'w')

            overwrite_remote_index = check_local_remote_sync(local_file, remote_session, flag)

    else:
        if remote_uuid:
            ## Init with the remote bytes - keeps the remote uuid and timestamp
            local_file = booklet.open(local_file_path, flag='n', init_bytes=remote_session._init_bytes)
            local_file._n_keys = 0

            overwrite_remote_index = True
        else:
            ## Else create a new file
            local_file = booklet.open(local_file_path, flag='n', key_serializer='str', value_serializer=value_serializer, n_buckets=n_buckets, buffer_size=buffer_size)

            overwrite_remote_index = True

    return local_file, overwrite_remote_index


def get_remote_index_file(local_file_path, overwrite_remote_index, remote_session, flag):
    """

    """
    remote_index_path = local_file_path.parent.joinpath(local_file_path.name + '.remote_index')

    if (not remote_index_path.exists() or overwrite_remote_index) and (flag != 'n'):
        if remote_session.uuid:
            index0 = remote_session.get_object()
            if index0.status == 200:
                with portalocker.Lock(remote_index_path, 'wb', timeout=120) as f:
                    f.write(index0.data)
            elif index0.status != 404:
                raise urllib3.exceptions.HTTPError(index0.error)

    return remote_index_path


def get_remote_value(local_file, key, remote_session):
    """

    """
    s3_key = '_metadata' if key == metadata_key_str else key
    resp = remote_session.get_object(s3_key)

    if resp.status == 200:
        timestamp = int(resp.metadata['timestamp'])

        if key == metadata_key_str:
            local_file.set_metadata(orjson.loads(resp.data), timestamp=timestamp)
        else:
            local_file.set(key, resp.data, timestamp, encode_value=False)
    elif resp.status == 404:
        pass
    else:
        return resp.error

    return None


def upload_group(group_id, local_file, remote_session, keys_in_group):
    entries = []
    for key in keys_in_group:
        result = local_file.get_timestamp(key, include_value=True, decode_value=False)
        if result is None:
            continue
        time_int_us, valb = result
        entries.append((key, time_int_us, valb))

    if entries:
        packed, offsets = pack_group(entries)
        resp = remote_session.put_object(str(group_id), packed)
        if resp.status // 100 != 2:
            return resp.error, None
        return None, offsets
    else:
        resp = remote_session.delete_object(str(group_id))
        return None, {}


def get_remote_group(group_id, local_file, remote_session):
    resp = remote_session.get_object(str(group_id))
    if resp.status == 200:
        entries = unpack_group(resp.data)
        for key, timestamp, value in entries:
            local_file.set(key, value, timestamp, encode_value=False)
    elif resp.status == 404:
        pass
    else:
        return resp.error

    return None


def get_remote_group_value(group_id, key, offset, length, timestamp_int, local_file, remote_session):
    resp = remote_session.get_object(str(group_id), range_start=offset, range_end=offset + length - 1)
    if resp.status in (200, 206):
        local_file.set(key, resp.data, timestamp_int, encode_value=False)
    elif resp.status == 404:
        pass
    else:
        return resp.error
    return None


def get_remote_group_values(group_id, key_infos, local_file, remote_session):
    """key_infos: list of (key, offset, length, timestamp_int)"""
    sorted_infos = sorted(key_infos, key=lambda x: x[1])
    range_start = sorted_infos[0][1]
    last = sorted_infos[-1]
    range_end = last[1] + last[2] - 1

    resp = remote_session.get_object(str(group_id), range_start=range_start, range_end=range_end)
    if resp.status in (200, 206):
        for key, offset, length, timestamp_int in sorted_infos:
            rel_offset = offset - range_start
            value = resp.data[rel_offset:rel_offset + length]
            local_file.set(key, value, timestamp_int, encode_value=False)
    elif resp.status == 404:
        pass
    else:
        return resp.error
    return None


def check_local_vs_remote(local_file, remote_time_bytes, key):
    """

    """
    if remote_time_bytes is None:
        return None

    remote_time_int = bytes_to_int(remote_time_bytes)
    local_time_int = local_file.get_timestamp(key)

    if local_time_int:
        if remote_time_int <= local_time_int:
            return False

    return True


#################################################
### local/remote changelog


def create_changelog(local_file_path, local_file, remote_index, remote_session):
    """
    Only check and save by the microsecond timestamp. Might need to add in the md5 hash if this is not sufficient.
    """
    changelog_path = local_file_path.parent.joinpath(local_file_path.name + '.changelog')
    if remote_index is not None:
        n_buckets = remote_index._n_buckets
    else:
        n_buckets = local_file._n_buckets

    with booklet.FixedLengthValue(changelog_path, 'n', key_serializer='str', value_len=14, n_buckets=n_buckets) as f:
        if remote_session.uuid and remote_index is not None:
            for key, local_int_us in local_file.timestamps():
                remote_val = remote_index.get(key)
                if remote_val:
                    remote_ts_bytes = remote_val[:7]
                    remote_int_us = bytes_to_int(remote_ts_bytes)
                    if local_int_us > remote_int_us:
                        local_bytes_us = int_to_bytes(local_int_us, 7)
                        f[key] = local_bytes_us + remote_ts_bytes
                else:
                    local_bytes_us = int_to_bytes(local_int_us, 7)
                    f[key] = local_bytes_us + int_to_bytes(0, 7)

            # Metadata
            key = booklet.utils.metadata_key_bytes.decode()
            local_int_us = local_file.get_timestamp(key)
            if local_int_us:
                remote_val = remote_index.get(key)
                if remote_val:
                    remote_ts_bytes = remote_val[:7]
                    remote_int_us = bytes_to_int(remote_ts_bytes)
                    local_int_us = local_file.get_timestamp(key)
                    if local_int_us > remote_int_us:
                        local_bytes_us = int_to_bytes(local_int_us, 7)
                        f[key] = local_bytes_us + remote_ts_bytes
                else:
                    local_bytes_us = int_to_bytes(local_int_us, 7)
                    f[key] = local_bytes_us + int_to_bytes(0, 7)
        else:
            for key, local_int_us in local_file.timestamps():
                local_bytes_us = int_to_bytes(local_int_us, 7)
                f[key] = local_bytes_us + int_to_bytes(0, 7)

            # Metadata
            key = booklet.utils.metadata_key_bytes.decode()
            local_int_us = local_file.get_timestamp(key)
            if local_int_us:
                local_bytes_us = int_to_bytes(local_int_us, 7)
                f[key] = local_bytes_us + int_to_bytes(0, 7)

    return changelog_path


def view_changelog(changelog_path):
    """

    """
    with booklet.FixedLengthValue(changelog_path) as f:
        for key, val in f.items():
            local_bytes_us = val[:7]
            remote_bytes_us = val[7:]
            local_int_us = bytes_to_int(local_bytes_us)
            remote_int_us = bytes_to_int(remote_bytes_us)
            if remote_int_us == 0:
                remote_ts = None
            else:
                remote_ts = datetime.fromtimestamp(remote_int_us*0.000001, tz=timezone.utc)

            dict1 = {
                'key': key,
                'remote_timestamp': remote_ts,
                'local_timestamp': datetime.fromtimestamp(local_int_us*0.000001, tz=timezone.utc)
                }

            yield dict1


##############################################
### Update remote


def upload_value(key, local_file, remote_session):
    """

    """
    time_int_us, valb = local_file.get_timestamp(key, include_value=True, decode_value=False)
    s3_key = '_metadata' if key == metadata_key_str else key
    resp = remote_session.put_object(s3_key, valb, {'timestamp': str(time_int_us)})
    if resp.status // 100 != 2:
        return resp.error
    else:
        return None


def update_remote(local_file, remote_index, changelog_path, remote_session, force_push, deletes, flag, ebooklet_type, num_groups=None):
    """

    """
    ## If file was open for replacement (n), then delete everything in the remote
    if flag == 'n':
        remote_session.delete_remote()

    ## Upload data and update the remote_index file
    updated = False
    failures = {}

    with booklet.FixedLengthValue(changelog_path) as cl:
        if num_groups is not None:
            ## Grouped upload path
            affected_group_ids = set()
            metadata_in_changelog = False

            for key in cl:
                if key == metadata_key_str:
                    metadata_in_changelog = True
                else:
                    affected_group_ids.add(key_to_group_id(key, num_groups))

            ## Also include groups affected by deletes
            for key in deletes:
                affected_group_ids.add(key_to_group_id(key, num_groups))

            ## Build full key lists per affected group from local_file
            group_keys = {gid: [] for gid in affected_group_ids}
            for key in local_file.keys():
                gid = key_to_group_id(key, num_groups)
                if gid in group_keys:
                    group_keys[gid].append(key)

            with ThreadPoolExecutor(max_workers=remote_session.threads) as executor:
                futures = {}

                ## Upload metadata individually if changed
                if metadata_in_changelog:
                    f = executor.submit(upload_value, metadata_key_str, local_file, remote_session)
                    futures[f] = ('metadata', metadata_key_str)

                ## Upload affected groups
                for gid, keys_in_group in group_keys.items():
                    f = executor.submit(upload_group, gid, local_file, remote_session, keys_in_group)
                    futures[f] = ('group', gid)

                for future in as_completed(futures):
                    tag, identifier = futures[future]
                    run_result = future.result()
                    if tag == 'metadata':
                        if run_result is None:
                            remote_index[metadata_key_str] = cl[metadata_key_str][:7] + b'\x00' * 8
                            updated = True
                        else:
                            failures[identifier] = run_result
                    else:
                        error, offsets = run_result
                        if error is None:
                            gid = identifier
                            for key in group_keys[gid]:
                                if key in offsets:
                                    ts = local_file.get_timestamp(key)
                                    if ts:
                                        offset, length = offsets[key]
                                        remote_index[key] = int_to_bytes(ts, 7) + int_to_bytes(offset, 4) + int_to_bytes(length, 4)
                            updated = True
                        else:
                            failures[identifier] = error

            ## Deletions are handled by group re-packing (deleted keys excluded)
            ## Remove deleted keys from remote_index
            for key in deletes:
                if key in remote_index:
                    del remote_index[key]
            deletes.clear()

        else:
            ## Per-key upload path (legacy)
            with ThreadPoolExecutor(max_workers=remote_session.threads) as executor:
                futures = {}
                for key in cl:
                    f = executor.submit(upload_value, key, local_file, remote_session)
                    futures[f] = key

                for future in as_completed(futures):
                    key = futures[future]
                    run_result = future.result()
                    if run_result is None:
                        remote_index[key] = cl[key][:7] + b'\x00' * 8
                        updated = True
                    else:
                        failures[key] = run_result

    if failures:
        logger.warning(f"There were {len(failures)} items that failed to upload. Please run this again.")

    ## Upload the remote_index file
    remote_index.sync()

    if updated or force_push or (deletes and num_groups is None):
        time_int_us = booklet.utils.make_timestamp_int()

        ## Get main file init bytes
        local_file._set_file_timestamp(time_int_us)
        local_file._file.seek(0)
        local_init_bytes = bytearray(local_file._file.read(200))
        if local_init_bytes[:16] != booklet.utils.uuid_variable_blt:
            raise ValueError(local_init_bytes)

        n_keys_pos = booklet.utils.n_keys_pos
        local_init_bytes[n_keys_pos:n_keys_pos+4] = b'\x00\x00\x00\x00'

        remote_index._file.seek(0)

        metadata = {
            'timestamp': str(time_int_us),
            'uuid': local_file.uuid.hex,
            'type': ebooklet_type,
            'init_bytes': base64.urlsafe_b64encode(local_init_bytes).decode(),
        }
        if num_groups is not None:
            metadata['num_groups'] = str(num_groups)

        resp = remote_session.put_db_object(remote_index._file.read(), metadata=metadata)

        if resp.status // 100 != 2:
            raise urllib3.exceptions.HTTPError("The db object failed to upload. You need to rerun the push with force_push=True or the remote will be corrupted.")

        ## remove deletes in remote (only for legacy per-key mode)
        if deletes and num_groups is None:
            remote_session.delete_objects(deletes)
            deletes.clear()

        updated = True

    if failures:
        return failures
    else:
        return updated


def indirect_copy_remote(source_session, target_session, source_key, target_key, source_bucket, dest_bucket):
    """

    """
    source_resp = source_session.get_object(source_key)

    if source_resp.status // 100 != 2:
        return source_resp

    target_resp = target_session.put_object(target_key, source_resp.stream, source_resp.metadata)

    return target_resp




















































